#ifndef __IFX_PPA_DIRECTPATH_H
#define __IFX_PPA_DIRECTPATH_H


/******************************************************************************
**
** FILE NAME    : ifx_ppa_directpath.h
** PROJECT      : Amazon-SE
** MODULES      : PPA 
**
** DATE         : 11 FEB 2008
** AUTHOR       : Santosh Gowda 
** DESCRIPTION  : PPA Directpath Header File
** COPYRIGHT    :   Copyright (c) 2006
**          Infineon Technologies AG
**          Am Campeon 1-12, 85579 Neubiberg, Germany
**
**   Any use of this software is subject to the conclusion of a respective
**   License agreement. Without such a License agreement no rights to the
**   software are granted
**
** HISTORY
** $Date        $Author         $Comment
** 11 FEB 2008  Santosh Gowda   Initiate Version
*******************************************************************************/

/*
 *  Chip Specific Head File
 */
#if defined(CONFIG_DANUBE_PPA) || defined(CONFIG_DANUBE_PPA_MODULE)
  #include <asm/danube/ifx_ppa_interface.h>
#elif defined(CONFIG_IFX_PPA) || defined(CONFIG_IFX_PPA_MODULE)
  #include <asm/amazon_se/ifx_ppa_interface.h>
#endif


typedef struct {
    uint32_t    tx_pkts;
    uint32_t    rx_pkts;
    uint32_t    tx_drop;
    uint32_t    rx_drop;
    uint32_t    tx_non_accel;
    uint32_t    rx_non_accel;

    int16_t	occupied;
    uint32_t	flags;

    PPA_NETIF   	*pDev;
    PPA_DIRECTPATH_CB 	*pDirectpathCb;

} PPA_DIRECTPATH_INFO;

/*
 * ####################################
 *             Declaration
 * ####################################
 */

/* Init/Exit */
int32_t ifx_ppa_directpath_init(uint32_t flags);
int32_t ifx_ppa_directpath_exit(uint32_t flags);
int32_t ifx_ppa_directpath_register_dev(uint32_t *if_id, PPA_NETIF *dev, PPA_DIRECTPATH_CB *pDirectpathCb, uint32_t flags);

/* Core functions */
int32_t ifx_ppa_directpath_rx (PPA_BUF *ppa_buf, int32_t len, uint32_t rx_if_id, uint32_t tx_if_mask, void *ppe_header);
int32_t ifx_ppa_directpath_tx (uint32_t rx_if_id, PPA_BUF *buf, int32_t len, uint32_t flags);
int32_t ifx_ppa_directpath_pause_tx (uint32_t if_id, PPA_FP_STOP_TX_FN stop_tx_fn, uint32_t flags);
int32_t ifx_ppa_directpath_resume_tx (uint32_t if_id, PPA_FP_RESTART_TX_FN start_tx_fn, uint32_t flags);
int32_t ifx_ppa_directpath_pause_rx (uint32_t if_id, uint32_t flags);
int32_t ifx_ppa_directpath_resume_rx (uint32_t if_id, uint32_t flags);

/* utility functions */
int32_t ifx_ppa_get_ifid (void);
PPA_NETIF *ifx_ppa_get_netif_for_ifid(uint32_t if_id);
int32_t ifx_get_ppa_ifid_for_netif (PPA_NETIF *dev);
#endif

